/**
 *  Hash Function Reference Header
 *  David Wilson, 2008
 */

/* Values from NIST API */
typedef unsigned char BitSequence;
typedef unsigned long long DataLength;
typedef enum {SUCCESS = 0,  FAIL = 1, BAD_HASHBITLEN = 2} HashReturn;

typedef struct {
  int hashbitlen;
  int numUnprocessed;        //Number of unprocessed bits
  BitSequence *unprocessed;  //Unprocessed data storage
  BitSequence *curr;         //Current result so far.
  DataLength datalen;        //Total length processed so far.

  //the following are used to generate the square-free sequence
  BitSequence *p[3];
  BitSequence *top[3];
  BitSequence parity;
  BitSequence small;
  BitSequence count;
  BitSequence move;

} hashState;

//state is where the initialized state gets returned
HashReturn Init(hashState *state, int hashbitlen);

//process as much data as you can, update state
HashReturn Update(hashState *state, const BitSequence *data, 
		  DataLength databitlen);

//perform final output filtering etc., return result in hashval
HashReturn Final(hashState *state, BitSequence *hashval);

//all-at-once hash - should call above fns
HashReturn Hash(int hashbitlen, const BitSequence *data, 
		DataLength databitlen, BitSequence *hashval);




//Primitive polynomial over GF(2^8) to use
#define PRIMITIVE 0x11d

//Block length
#define BLOCK_LENGTH_BYTES 64
#define PADDED_BLOCK_LENGTH_BITS 512

//S-box transform.
//This is inv(x) xor 3

static const BitSequence SBOX[256] = 
  {0x03, 0x02, 0x8D, 0xF7, 0x44, 0xA4, 0x79, 0xB9, 
   0xAE, 0x9E, 0xDE, 0x9B, 0x3E, 0xA9, 0x5E, 0x95, 
   0xDB, 0x71, 0xC3, 0x5B, 0xE3, 0x3D, 0x4F, 0x65, 
   0x93, 0xDD, 0x56, 0x83, 0xA3, 0x80, 0x48, 0x29, 
   0x6F, 0xEE, 0x3A, 0x52, 0x63, 0x55, 0x2F, 0x89, 
   0x73, 0xD3, 0x1C, 0x49, 0x25, 0x88, 0x30, 0x6D, 
   0x4B, 0x8A, 0x6C, 0x2D, 0xA7, 0xC0, 0x43, 0x5D, 
   0x53, 0x21, 0xCC, 0xAA, 0xA8, 0x0F, 0x16, 0xE2, 

   0x35, 0x5C, 0xFB, 0xD6, 0x91, 0x4D, 0xA5, 0x07, 
   0x33, 0x8B, 0x28, 0x1D, 0x15, 0x64, 0x46, 0x90, 
   0x3B, 0x20, 0x6B, 0x8F, 0x82, 0x19, 0x26, 0x62, 
   0x10, 0xC2, 0xC8, 0x60, 0x94, 0x0D, 0x34, 0x42, 
   0x27, 0x54, 0xC9, 0x58, 0xBA, 0xC7, 0x14, 0x4E, 
   0x51, 0x8E, 0xEC, 0xB0, 0x23, 0xEF, 0x2C, 0x31, 
   0x2B, 0xD2, 0x12, 0xDA, 0xEA, 0xF8, 0xD9, 0x7A, 
   0xD8, 0x74, 0x05, 0xB8, 0x87, 0xCE, 0xFD, 0xFF, 

   0x18, 0x57, 0xA2, 0x1E, 0x7F, 0xCF, 0xE7, 0xB3, 
   0x4A, 0x32, 0x24, 0x2E, 0x50, 0x6A, 0x01, 0xF6, 
   0x1B, 0xDC, 0x47, 0x4C, 0x98, 0xBF, 0x0C, 0x5F, 
   0x08, 0xDF, 0xBE, 0x97, 0xAF, 0x0A, 0xC4, 0xA1, 
   0x1F, 0x81, 0x9C, 0xC5, 0x37, 0xC1, 0x45, 0x06, 
   0xCD, 0x38, 0x0E, 0x3F, 0x9F, 0x0B, 0xBD, 0xB4, 
   0x84, 0xE6, 0xED, 0x68, 0xE8, 0xF1, 0xBC, 0xAC, 
   0xC6, 0x67, 0x04, 0x78, 0x96, 0x99, 0xAD, 0xB5, 

   0x11, 0x5A, 0xA6, 0x36, 0x66, 0xBB, 0xA0, 0x9D, 
   0xD1, 0xF4, 0x61, 0x59, 0x86, 0x7E, 0xAB, 0x39, 
   0x2A, 0x72, 0xCB, 0xF5, 0xFA, 0x40, 0xD4, 0xD5, 
   0x13, 0x70, 0x75, 0x7B, 0x9A, 0x09, 0x1A, 0x92, 
   0x17, 0x3C, 0xE5, 0xF3, 0x85, 0xB2, 0xE1, 0xF2, 
   0xF9, 0x77, 0xF0, 0xB7, 0x6E, 0x22, 0xB1, 0x69, 
   0xE0, 0xE4, 0xB6, 0xE9, 0x00, 0x8C, 0xD0, 0xCA, 
   0x41, 0xD7, 0xEB, 0x76, 0x7C, 0xFC, 0x7D, 0xFE};


static const BitSequence *ROUND_KEY[4] = 
  {&SBOX[0], &SBOX[64], &SBOX[128], &SBOX[192]};


//Galois field GF(2^8).  gf[i] = a^i in vector form.  Zero element not included.
static const BitSequence gf[256] = 
  {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,
   0x1D, 0x3A, 0x74, 0xE8, 0xCD, 0x87, 0x13, 0x26,
   0x4C, 0x98, 0x2D, 0x5A, 0xB4, 0x75, 0xEA, 0xC9,
   0x8F, 0x03, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0,
   0x9D, 0x27, 0x4E, 0x9C, 0x25, 0x4A, 0x94, 0x35,
   0x6A, 0xD4, 0xB5, 0x77, 0xEE, 0xC1, 0x9F, 0x23,
   0x46, 0x8C, 0x05, 0x0A, 0x14, 0x28, 0x50, 0xA0,
   0x5D, 0xBA, 0x69, 0xD2, 0xB9, 0x6F, 0xDE, 0xA1,

   0x5F, 0xBE, 0x61, 0xC2, 0x99, 0x2F, 0x5E, 0xBC,
   0x65, 0xCA, 0x89, 0x0F, 0x1E, 0x3C, 0x78, 0xF0,
   0xFD, 0xE7, 0xD3, 0xBB, 0x6B, 0xD6, 0xB1, 0x7F,
   0xFE, 0xE1, 0xDF, 0xA3, 0x5B, 0xB6, 0x71, 0xE2,
   0xD9, 0xAF, 0x43, 0x86, 0x11, 0x22, 0x44, 0x88,
   0x0D, 0x1A, 0x34, 0x68, 0xD0, 0xBD, 0x67, 0xCE,
   0x81, 0x1F, 0x3E, 0x7C, 0xF8, 0xED, 0xC7, 0x93,
   0x3B, 0x76, 0xEC, 0xC5, 0x97, 0x33, 0x66, 0xCC,

   0x85, 0x17, 0x2E, 0x5C, 0xB8, 0x6D, 0xDA, 0xA9,
   0x4F, 0x9E, 0x21, 0x42, 0x84, 0x15, 0x2A, 0x54,
   0xA8, 0x4D, 0x9A, 0x29, 0x52, 0xA4, 0x55, 0xAA,
   0x49, 0x92, 0x39, 0x72, 0xE4, 0xD5, 0xB7, 0x73,
   0xE6, 0xD1, 0xBF, 0x63, 0xC6, 0x91, 0x3F, 0x7E,
   0xFC, 0xE5, 0xD7, 0xB3, 0x7B, 0xF6, 0xF1, 0xFF,
   0xE3, 0xDB, 0xAB, 0x4B, 0x96, 0x31, 0x62, 0xC4,
   0x95, 0x37, 0x6E, 0xDC, 0xA5, 0x57, 0xAE, 0x41,

   0x82, 0x19, 0x32, 0x64, 0xC8, 0x8D, 0x07, 0x0E,
   0x1C, 0x38, 0x70, 0xE0, 0xDD, 0xA7, 0x53, 0xA6,
   0x51, 0xA2, 0x59, 0xB2, 0x79, 0xF2, 0xF9, 0xEF,
   0xC3, 0x9B, 0x2B, 0x56, 0xAC, 0x45, 0x8A, 0x09,
   0x12, 0x24, 0x48, 0x90, 0x3D, 0x7A, 0xF4, 0xF5,
   0xF7, 0xF3, 0xFB, 0xEB, 0xCB, 0x8B, 0x0B, 0x16,
   0x2C, 0x58, 0xB0, 0x7D, 0xFA, 0xE9, 0xCF, 0x83,
   0x1B, 0x36, 0x6C, 0xD8, 0xAD, 0x47, 0x8E, 0x01};

//Inverse mapping of GF(2^8).  gfinv[i] = j s.t. a^(j) = gf[j] = i.
static const BitSequence gfinv[256] = 
  {0x00, 0x00, 0x01, 0x19, 0x02, 0x32, 0x1A, 0xC6,
   0x03, 0xDF, 0x33, 0xEE, 0x1B, 0x68, 0xC7, 0x4B,
   0x04, 0x64, 0xE0, 0x0E, 0x34, 0x8D, 0xEF, 0x81,
   0x1C, 0xC1, 0x69, 0xF8, 0xC8, 0x08, 0x4C, 0x71,
   0x05, 0x8A, 0x65, 0x2F, 0xE1, 0x24, 0x0F, 0x21,
   0x35, 0x93, 0x8E, 0xDA, 0xF0, 0x12, 0x82, 0x45,
   0x1D, 0xB5, 0xC2, 0x7D, 0x6A, 0x27, 0xF9, 0xB9,
   0xC9, 0x9A, 0x09, 0x78, 0x4D, 0xE4, 0x72, 0xA6,

   0x06, 0xBF, 0x8B, 0x62, 0x66, 0xDD, 0x30, 0xFD,
   0xE2, 0x98, 0x25, 0xB3, 0x10, 0x91, 0x22, 0x88,
   0x36, 0xD0, 0x94, 0xCE, 0x8F, 0x96, 0xDB, 0xBD,
   0xF1, 0xD2, 0x13, 0x5C, 0x83, 0x38, 0x46, 0x40,
   0x1E, 0x42, 0xB6, 0xA3, 0xC3, 0x48, 0x7E, 0x6E,
   0x6B, 0x3A, 0x28, 0x54, 0xFA, 0x85, 0xBA, 0x3D,
   0xCA, 0x5E, 0x9B, 0x9F, 0x0A, 0x15, 0x79, 0x2B,
   0x4E, 0xD4, 0xE5, 0xAC, 0x73, 0xF3, 0xA7, 0x57,

   0x07, 0x70, 0xC0, 0xF7, 0x8C, 0x80, 0x63, 0x0D,
   0x67, 0x4A, 0xDE, 0xED, 0x31, 0xC5, 0xFE, 0x18,
   0xE3, 0xA5, 0x99, 0x77, 0x26, 0xB8, 0xB4, 0x7C,
   0x11, 0x44, 0x92, 0xD9, 0x23, 0x20, 0x89, 0x2E,
   0x37, 0x3F, 0xD1, 0x5B, 0x95, 0xBC, 0xCF, 0xCD,
   0x90, 0x87, 0x97, 0xB2, 0xDC, 0xFC, 0xBE, 0x61,
   0xF2, 0x56, 0xD3, 0xAB, 0x14, 0x2A, 0x5D, 0x9E,
   0x84, 0x3C, 0x39, 0x53, 0x47, 0x6D, 0x41, 0xA2,

   0x1F, 0x2D, 0x43, 0xD8, 0xB7, 0x7B, 0xA4, 0x76,
   0xC4, 0x17, 0x49, 0xEC, 0x7F, 0x0C, 0x6F, 0xF6,
   0x6C, 0xA1, 0x3B, 0x52, 0x29, 0x9D, 0x55, 0xAA,
   0xFB, 0x60, 0x86, 0xB1, 0xBB, 0xCC, 0x3E, 0x5A,
   0xCB, 0x59, 0x5F, 0xB0, 0x9C, 0xA9, 0xA0, 0x51,
   0x0B, 0xF5, 0x16, 0xEB, 0x7A, 0x75, 0x2C, 0xD7,
   0x4F, 0xAE, 0xD5, 0xE9, 0xE6, 0xE7, 0xAD, 0xE8,
   0x74, 0xD6, 0xF4, 0xEA, 0xA8, 0x50, 0x58, 0xAF};


void setNextSequenceValue(hashState *state);
void setLastSequenceValue(hashState *state);
void hashOneBlock(hashState *state);
void addInto(BitSequence *to, const BitSequence *from);
void doSBox(BitSequence *data);
void doTransform(BitSequence *data);

//Number of rounds
#define NUM_ROUNDS 4

